from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
m3u8 = LazyLoader("m3u8", globals())
rich = LazyLoader("rich", globals())
columns = LazyLoader("columns", globals(), "rich.columns")
panel = LazyLoader("panel", globals(), "rich.panel")

utils = LazyLoader("utils", globals(), "pyxk.utils")



class M3U8Parse:
    """M3U8 解析类"""

    def __init__(
        self, *, content: str, url: str, m3u8_instance: object
    ):
        self.obj = m3u8.loads(content=content, uri=url)
        self.m3u = m3u8_instance
        self.url = url
        self.keys = None
        self.segments = None
        self.maximum  = None
        self.duration = None


    @classmethod
    def start_parsing_from_content(
        cls, *, content: str, url: str, m3u8_instance: object
    ):
        """类方法 解析 m3u8"""
        self = cls(content=content, url=url, m3u8_instance=m3u8_instance)
        self.start_parsing()
        return self


    def start_parsing(self):
        """解析 m3u8 入口"""

        self.parsing_playlists()
        self.parsing_keys()
        self.parsing_segments()
        self.print_parsing_result()


    def parsing_playlists(self):
        """解析 m3u8 playlists"""

        if not self.obj.is_variant:
            return

        # 排序 playlists, 获取最大带宽 m3u8 链接
        def _sorted(playlist):
            playlist.uri = playlist.absolute_uri
            return playlist.stream_info.bandwidth
        playlists = sorted(self.obj.playlists, key=_sorted)

        # 保存 m3u8 文件
        self.m3u.sava_m3u8_content(url=self.url, content=self.obj.dumps())

        # 加载 playlist 数据
        self.url = variant_url = playlists[-1].uri
        variant_content = self.m3u.get_m3u8_content(url=variant_url)
        self.obj = m3u8.loads(content=variant_content, uri=variant_url)


    def parsing_keys(self):
        """解析 m3u8 keys"""

        m3u8_keys = {}
        for key in self.obj.keys:
            if not key:
                continue
            url = key.uri = key.absolute_uri
            secret = self.m3u.get_m3u8_content(url=url, is_m3u8key=True)
            m3u8_keys[url] = {
                "key": secret,
                "iv": key.iv.removeprefix("0x")[:16] if key.iv else secret[:16]
            }
            # 保存密钥
            self.m3u.sava_m3u8_content(url=url, content=secret, is_m3u8key=True)
        self.keys = m3u8_keys


    def parsing_segments(self):
        """解析 m3u8 segments"""

        segments_dict, duration = {}, 0
        for index, segment in enumerate(self.obj.segments):
            url = segment.uri = segment.absolute_uri
            key, duration = None, segment.duration + duration

            # segment 加密 - 解析
            if segment.key and self.keys:
                if segment.key.uri in self.keys:
                    key = segment.key.uri
            segments_dict[index] = {"url": url, "key": key}

        # 保存 m3u8 文件
        if self.obj.is_endlist:
            self.m3u.sava_m3u8_content(url=self.url, content=self.obj.dumps())

        self.segments = segments_dict
        self.duration = duration
        self.maximum  = len(segments_dict)


    def print_parsing_result(self):
        """打印 m3u8 解析结果"""

        filepath, filename = os.path.split(self.m3u.output) if self.m3u.output else (None, None)
        m3u8_url = panel.Panel(
            columns.Columns([self.url if self.url else "None"], expand=True, align="center")
        )

        information = [
            f"Limit: {self.m3u.limit}",
            f"Maximum: {self.maximum}",
            f"PlayTime: {utils.human_playtime_pr(self.duration)}",
            f"Encryption: {bool(self.keys)}",
            f"FileName: {filename}",
            f"FilePath: {filepath}",
            "Result: [yellow]==>[/] " \
            + ("[green b]Parsing success ![/]" \
            if self.obj.is_endlist \
            else "[red b]Parsing failure ![/]")
        ]

        information = [f"{i+1} {x}" for i, x in enumerate(information)]
        information.insert(0, m3u8_url)
        information = panel.Panel(
            columns.Columns(information, expand=True),
            border_style="yellow"
        )

        rich.print(information)


    def close(self):
        """关闭"""
        self.obj = None
        self.m3u = None
        self.url = None
        self.keys = None
        self.segments = None
        self.maximum  = None
        self.duration = None



from pyxk.m3u8._parser import M3U8Parse
from pyxk.m3u8._download import SegmentsDownload
from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
utils = LazyLoader("utils", globals(), "pyxk.utils")
requests = LazyLoader("requests", globals(), "pyxk.requests")



class BaseM3U8:
    """Base M3U8"""

    default_limit = 8

    def __init__(self):
        # m3u8 视频路径
        self._output   = None
        # 下载完成, 删除 m3u8 文件
        self._delete   = False
        # 下载完成, 保留 segments 文件
        self._reserve  = False
        # request.Session()
        self._session  = requests.Session()
        # headers
        self._headers = None
        # verify
        self._verify  = True
        # 不使用本地 m3u8 文件, 重新请求
        self._rerequest = False
        # 异步请求 limit
        self._limit = self.default_limit
        # 所有保存文件的路径
        self.result = []

    @property
    def output(self):
        """m3u8视频文件路径, output=None不下载"""
        return self._output

    @output.setter
    def output(self, path):
        if not path or not isinstance(path, str):
            self._output = None
        else:
            # 去除 output 中间的空白字符
            path = "-".join(path.split())
            # 得到绝对路径
            path = os.path.normpath(os.path.abspath(path))
            # 文件路径去重
            path = utils.rename_file(path, suffix="mp4")[0]
            self._output = path

    @property
    def m3u8temp(self):
        """m3u8文本文件 保存路径(文件夹)"""
        if not self.output:
            return None
        dirname, basename = os.path.split(self.output)
        basename = "." + basename.removesuffix(".mp4") + ".temp"
        return os.path.join(dirname, basename)

    @property
    def limit(self):
        """异步请求 limit"""
        return self._limit

    @limit.setter
    def limit(self, limit):
        if not isinstance(limit, int) or limit <= 0:
            limit = self.default_limit
        self._limit = limit


    def get_m3u8_content(
        self, url: str, *, is_m3u8key: bool=False
    ):
        """获取 m3u8 文件内容

        :params: url: m3u8 链接
        :params: is_m3u8key: 链接是否是 m3u8密钥 链接

        从 url 链接 或者 本地文件获取m3u8文本内容
        若 rerequests=True or self.m3u8temp为假, 则始终从 url 获取内容
        """
        mode, suffix, encoding, attri = "r", ".m3u8", "utf-8", "text"
        if is_m3u8key:
            mode, suffix, encoding, attri = "rb", ".key", None, "content"

        # 文件完整路径
        file = os.path.join(self.m3u8temp, utils.hash256(url)+suffix) if self.m3u8temp else None
        # 从 url 获取 m3u8 内容
        if (
            self._rerequest
            or not file
            or not os.path.isfile(file)
        ):
            response = self._session.get(url=url, headers=self._headers, verify=self._verify)
            content  = getattr(response, attri)
        # 从本地获取 m3u8 内容
        else:
            with open(file=file, mode=mode, encoding=encoding) as fileobj:
                content = fileobj.read()
        return content


    def sava_m3u8_content(
        self, url: str, content: str, *, is_m3u8key: bool=False
    ):
        """保存 m3u8 内容

        :params: url: m3u8 链接
        :params: content: 保存内容
        :params: is_m3u8key: 链接是否是 m3u8密钥 链接

        保存 content 至本地文件
        filepath = self.m3u8temp, filename = hash256(url)
        若 url or self.m3u8temp 为假, 则不保存
        """
        # 若 url or self.m3u8temp 为假, 则不保存
        if (
            not url
            or not self.m3u8temp
            or not isinstance(url, str)
        ):
            return

        mode, suffix, encoding = "w", ".m3u8", "utf-8"
        if is_m3u8key:
            mode, suffix, encoding = "wb", ".key", None

        file = os.path.join(self.m3u8temp, utils.hash256(url)+suffix)
        self.result.append(file)
        # self._rerequests = True or 本地文件不存在 才保存
        if self._rerequest or not os.path.isfile(file):
            with utils.open(file=file, mode=mode, encoding=encoding) as fileobj:
                fileobj.write(content)


class M3U8(BaseM3U8):
    """ Load M3U8 """

    def load_url(self, url: str, *, output: str=None, **kwargs):
        """负载 m3u8 链接

        :params: url: m3u8 链接
        :params: output: 视频输出
        """
        self._initialization_attri(output=output, **kwargs)
        content = self.get_m3u8_content(url=url)
        self._call(content=content, url=url)


    def load_content(self, content: str, url: str=None, *, output: str=None, **kwargs):
        """负载 m3u8 内容 或 文件路径

        :params: content: m3u8文本内容 或者 m3u8本地文件路径
        :params: url: m3u8 链接
        :params: output: 视频输出
        """
        self._initialization_attri(output=output, **kwargs)
        # m3u8文件规定必需以 '#EXTM3U' 开头
        if not content.startswith("#EXTM3U"):
            _m3u8file = os.path.normpath(os.path.abspath(content))
            if os.path.isfile(_m3u8file):
                with open(_m3u8file, "r", encoding="utf-8") as fileobj:
                    content = fileobj.read()
        self._call(content=content, url=url)


    def _call(self, *, content, url=None):
        """调用 M3U8Parse 进行解析"""

        m3u8parse = M3U8Parse.start_parsing_from_content(
            content=content, url=url, m3u8_instance=self
        )
        # 判断 output是否存在
        output_exists = lambda: os.path.isfile(self.output) if self.output else False
        # 开启 异步下载
        if self.output and m3u8parse.segments:
            download = SegmentsDownload.start_download_from_segments(m3u8parse=m3u8parse)
            # 删除 segments
            if not self._reserve and output_exists():
                self._remove(download.result, download.store)
        # 删除 m3u8 文件
        if self._delete and self.m3u8temp:
            self._remove(files=self.result, folder=self.m3u8temp)


    @staticmethod
    def _remove(files: list, folder: str):
        """删除文件和文件夹"""

        for file in files:
            os.remove(file)
        if not isinstance(folder, str):
            return
        if os.path.isdir(folder) and not os.listdir(folder):
            os.rmdir(folder)


    def _initialization_attri(
        self, output=None, rerequest=False, delete=False, reserve=False, headers=None, verify=True, limit=None
    ):
        # output
        self.output = output
        # rerequest
        self._rerequest = bool(rerequest)
        # delete
        self._delete = bool(delete)
        # reserve
        self._reserve = bool(reserve)
        # headers
        self._headers = dict(headers or {})
        # verify
        self._verify = bool(verify)
        # limit
        if limit is not None:
            self.limit = limit
        self.result = []


def load_url(url, *, output=None, rerequest=False, delete=False, reserve=False, headers=None, verify=True, limit=None):
    m3u8_obj = M3U8()
    return m3u8_obj.load_url(url=url, output=output, rerequest=rerequest, delete=delete, reserve=reserve, headers=headers, verify=verify, limit=limit)


def load_content(content, url=None, *, output=None, rerequest=False, delete=False, reserve=False, headers=None, verify=True, limit=None):
    m3u8_obj = M3U8()
    return m3u8_obj.load_content(content=content, url=url, output=output, rerequest=rerequest, delete=delete, reserve=reserve, headers=headers, verify=verify, limit=limit)



"""
import pyxk.utils
"""
from pyxk.lazy_loader import LazyLoader

re = LazyLoader("re", globals())
os = LazyLoader("os", globals())
time = LazyLoader("time", globals())
typing = LazyLoader("typing", globals())
base64 = LazyLoader("base64", globals())
hashlib = LazyLoader("hashlib", globals())
difflib = LazyLoader("difflib", globals())
warnings = LazyLoader("warnings", globals())
multidict = LazyLoader("multidict", globals())
functools = LazyLoader("functools", globals())
itertools = LazyLoader("itertools", globals())
collections = LazyLoader("collections", globals())



def runtime(func):
    """计算函数运行时间"""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        ret = func(*args, **kwargs)
        end_time   = time.perf_counter()
        try:
            from rich import print
        except ModuleNotFoundError:
            pass
        print(f"{func.__name__!r} runtime:", end_time - start_time)
        return ret
    return wrapper


def async_runtime(func):
    """计算异步函数运行时间"""
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        ret = await func(*args, **kwargs)
        end_time   = time.perf_counter()
        try:
            from rich import print
        except ModuleNotFoundError:
            pass
        print(f"{func.__name__!r} runtime:", end_time - start_time)
        return ret
    return wrapper


def open_decorator(func):
    """
    内置函数 open 装饰器
    作用: 写或追加模式下 创建不存在的目录
    """
    @functools.wraps(func)
    def wrapper(
        file,
        mode="r",
        buffering=-1,
        encoding=None,
        errors=None,
        newline=None,
        closefd=True,
        opener=None
    ):
        if not isinstance(mode, str):
            raise TypeError(
                f"{func.__name__}() argument 'mode'"
                f"must be str, not {type(mode).__name__!r}"
            )

        # 判断 mode 是否属于 写或追加 模式
        # collections.Counter 统计可迭代对象 每项出现的次
        # itertools.product 求多个可迭代对象的笛卡尔积
        _mode = tuple(
            collections.Counter(i+j)
            for i in ("a", "w")
            for j in ("b", "+", "b+", "")
        )

        # 创建目录
        if collections.Counter(mode) in _mode:
            os.makedirs(os.path.dirname(file), exist_ok=True)
        # 二进制模式下不应该使用 encoding
        if mode.find("b") != -1 and encoding is not None:
            warnings.warn(
                "binary mode doesn't take an encoding argument",
                DeprecationWarning
            )
            encoding = None
        return func(
            file=file,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener
        )
    return wrapper

make_open = open = open_decorator(open)


def is_base64(data: typing.Union[str, bytes]) -> bool:
    """判断base64数据类型

    :params: data: 需要检测的数据
    """
    if isinstance(data, bytes):
        # base64 数据类型 正则表达式判断
        B64_RE_PATTERN_B = re.compile(rb"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$")
        return bool(B64_RE_PATTERN_B.match(data))

    if isinstance(data, str):
        # base64 数据类型 正则表达式判断
        B64_RE_PATTERN   = re.compile(r"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$")
        return bool(B64_RE_PATTERN.match(data))
    # str 或 bytes 以外类型返回 False
    return False


def tobytes_from_base64(
    data: typing.Union[str, bytes], encoding="UTF-8"
) -> typing.Tuple[bool, typing.Union[str, bytes]]:
    """base64数据类型 转化为bytes

    :params: data: 需要 base64 解密的数据
    :params: encoding: type(data) is 'str' 通过 encoding 转换为 bytes
    """
    if (
        not isinstance(data, (str, bytes))
        or not is_base64(data)
    ):
        return False, data
    if isinstance(data, str):
        data = data.encode(encoding)
    return True, base64.b64decode(data)


# User-Agnet
UA_ANDROID  = "Mozilla/5.0 (Linux; Android 11; Pixel 5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.91 Mobile Safari/537.36"
UA_WINDOWNS = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36"
UA_MAC      = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36"
UA_IPHONE   = "Mozilla/5.0 (iPhone; CPU iPhone OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1"
UA_IPAD     = "Mozilla/5.0 (iPad; CPU OS 13_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/87.0.4280.77 Mobile/15E148 Safari/604.1"
UA_SYMBIAN  = "Mozilla/5.0 (Symbian/3; Series60/5.2 NokiaN8-00/012.002; Profile/MIDP-2.1 Configuration/CLDC-1.1 ) AppleWebKit/533.4 (KHTML, like Gecko) NokiaBrowser/7.3.0 Mobile Safari/533.4 3gpp-gba"
UA_APAD     = "Mozilla/5.0 (Linux; Android 11; Phh-Treble vanilla Build/RQ3A.211001.001;) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/90.0.4430.91 Safari/537.36"
UA_ALL = {
    "android" : UA_ANDROID,
    "windows" : UA_WINDOWNS,
    "mac"     : UA_MAC,
    "iphone"  : UA_IPHONE,
    "ipad"    : UA_IPAD,
    "symbian" : UA_SYMBIAN,
    "apad"    : UA_APAD
}


def get_user_agent(
    user_agent: str, overwrite: bool=False
) -> str:
    """获取 UserAgent

    :params: user_agent: 返回 useragent 模糊查找的结果
    :params: overwrite: 重写 useragent, 若为True, 直接返回useragent
    """
    if not isinstance(user_agent, str):
        raise ValueError(
            f"invalid user_agent: '{user_agent}', "
            f"type: '{type(user_agent).__name__}'")
    # 重写
    if overwrite:
        return user_agent

    user_agent = difflib.get_close_matches(user_agent.lower(), UA_ALL, 1)
    if not user_agent:
        return UA_ANDROID
    return UA_ALL[user_agent[0]]


def default_headers():
    """默认 headers

    headers = {"User-Agent": "..."}
    """
    headers = (("User-Agent", get_user_agent("android")),)
    return multidict.CIMultiDict(headers)


def md5(
    data: typing.Union[str, bytes], encoding: str="UTF-8"
) -> str:
    """ MD5 加密

    :params: data: 需要加密的数据
    :params: encoding: type(data) is 'str' 通过 encoding 转换为 bytes
    """
    if isinstance(data, str):
        data = data.encode(encoding=encoding)

    elif not isinstance(data, bytes):
        raise TypeError(
            "md5 encrypted data must be a 'str' or 'bytes'"
            f", can't be a {type(data).__name__!r}")
    return hashlib.md5(data).hexdigest()


def rename_file(file: str, *, suffix: str=None) -> typing.Tuple[str, str, str]:
    """重命名本地存在的文件
    :params: file: 文件路径
    :params: suffix: 文件后缀名

    return:
    Tuple(rename_file, dirname, basename)
    """
    # 绝对路径
    file = os.path.abspath(file)
    # 后缀名解析
    if not isinstance(suffix, str) or not suffix:
        file_split, suffix = file.rsplit(".", 1), ""
        if len(file_split) == 2 and len(file_split[-1]) < 6:
            suffix = file_split[-1]
    suffix = "." + suffix.strip().removeprefix(".") if suffix.strip() else suffix.strip()
    # 完整文件路径
    if not file.endswith(suffix):
        file += suffix

    if not os.path.isfile(file):
        return file, *os.path.split(file)

    for index in itertools.count(1):
        newfile = file.removesuffix(suffix) + f".{index}" + suffix
        if not os.path.isfile(newfile):
            break
    return newfile, *os.path.split(newfile)


def rename_folder(folder: str) -> typing.Tuple[str, str, str]:
    """重命名本地存在的文件夹
    :params: folder: 文件夹路径

    return:
    Tuple(rename_folder, dirname, basename)
    """
    folder = os.path.abspath(folder)

    if not os.path.isdir(folder):
        return folder, *os.path.split(folder)

    for index in itertools.count(1):
        new_folder = folder + f".{index}"
        if not os.path.isdir(new_folder):
            break
    return new_folder, *os.path.split(new_folder)


def human_playtime_pr(
    playtime: typing.Union[int, float]
) -> str:
    """人类直观时间展示

    :params: playtime: 传入一个时间(秒), 返回人类能理解的时间格式
    """
    if not isinstance(playtime, (int, float)):
        if playtime is None:
            return None
        raise TypeError(
            f"'playtime' must be a 'int' or 'float', got {type(playtime).__name__!r}")

    symbol, playtime = "-" if playtime < 0 else "", round(abs(playtime))
    hour, second = divmod(playtime, 3600)
    minute, second = divmod(second, 60)
    return f"{symbol}{hour}:{minute:0>2}:{second:0>2}"


def hash256(
    data:typing.Union[str, bytes], encoding: str="UTF-8"
) -> str:
    """HASH_256

    :params: data: 需要加密的数据
    :params: encoding: type(data) is 'str' 通过 encoding 转换为 bytes
    """
    if isinstance(data, str):
        data = data.encode(encoding)
    return hashlib.sha256(data).hexdigest()


def transfer_lazyloader(code: str, /) -> str:
    """python模块转换懒加载

    :params: code: python代码
    """
    def repl_string(match):

        module_name  = match.group("module_name")
        package_name = match.group("package_name")
        import_name  = match.group("import_name")

        if module_name:
            string = "\n".join(
                [
                    f'{item.strip()} = LazyLoader("{item.strip()}", globals(), "{package_name}.{item.strip()}")'
                    for item in module_name.split(",")
                ]
            )
        elif import_name:
            string = "\n".join(
                [
                    f'{item.strip()} = LazyLoader("{item.strip()}", globals())'
                    for item in import_name.split(",")
                ]
            )
        return string

    pattern = re.compile(r"^from\s+?(?P<package_name>.*?)\s+?import\s+?(?P<module_name>.*?)\s*?$|^import\s+?(?P<import_name>.*?)\s*?$", flags=re.M)
    lazy_loader = "from pyxk.lazy_loader import LazyLoader\n\n"
    return lazy_loader + pattern.sub(repl=repl_string, string=code)


def transfer_lazyloader_from_file(
    read_file: str, write_file: str, *, encoding="utf-8"
):
    """python模块转换懒加载 从文件

    :params: read_file: 读取python代码文件
    :params: write_file: 写入转换后的python代码文件
    """
    with open(read_file, "r", encoding=encoding) as fileobj:
        content = fileobj.read()

    with open(write_file, "w", encoding=encoding) as fileobj:
        content = transfer_lazyloader(content)
        fileobj.write(content)



from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
time = LazyLoader("time", globals())
shlex = LazyLoader("shlex", globals())
typing = LazyLoader("typing", globals())
warnings = LazyLoader("warnings", globals())
asyncio = LazyLoader("asyncio", globals())
threading = LazyLoader("threading", globals())
subprocess = LazyLoader("subprocess", globals())

aiohttp = LazyLoader("aiohttp", globals())
aiofiles = LazyLoader("aiofiles", globals())
progress = LazyLoader("progress", globals(), "rich.progress")

aes = LazyLoader("aes", globals(), "pyxk.aes")
utils = LazyLoader("utils", globals(), "pyxk.utils")



class SegmentsDownload:
    """Segments 异步下载"""

    def __init__(
        self, *, m3u8parse: typing.Union["M3U8Parse"]
    ):
        # m3u8 解析类
        self.m3u8parse = m3u8parse
        # m3u8 keys -> cipher
        self.cipher = {
            index: aes.Cryptor(**cipher)
            for index, cipher in m3u8parse.keys.items()
        } or None
        # segments 下载保存目录
        self.store = os.path.join(m3u8parse.m3u.m3u8temp, "segments")
        # limit
        self.limit = m3u8parse.m3u.limit
        # 异步请求默认超时时间
        self.default_timeout = 2 * 60
        # rich 进度条
        self.progress = None
        self.progress_task = None
        # 下载完成后全部文件路径
        self.result = []


    def close(self):
        """关闭"""
        self.m3u8parse.close()
        self.cipher = None
        self.progress = None
        self.progress_task = None


    def start_download(self):
        """异步下载 入口"""

        # 创建文件夹
        os.makedirs(self.store, exist_ok=True)
        # 开启下载 进度条
        with progress.Progress(
            *(
                progress.SpinnerColumn("line"),
                #progress.TextColumn("[progress.description]{task.description}"),
                progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                progress.BarColumn(finished_style="green"),
                progress.TextColumn("[progress.description]{task.completed}/{task.total}"),
                progress.TimeElapsedColumn(),
                # progress.TransferSpeedColumn(),
            ),
        ) as self.progress:
            self.progress_task = self.progress.add_task(
                description=None,
                total=self.m3u8parse.maximum
            )
            result = asyncio.run(self.gather_request())
        self.result = result
        self.merge_segments()


    async def gather_request(self):
        """异步下载管理器 收集所有异步任务"""

        headers   = utils.default_headers()
        timeout   = aiohttp.ClientTimeout(total=self.default_timeout)
        connector = aiohttp.TCPConnector(limit=self.limit)

        async with aiohttp.ClientSession(
            timeout=timeout, headers=headers, connector=connector
        ) as session:

            tasks = [
                self.request(session=session, file=file, **segment)
                for file, segment in self.m3u8parse.segments.items()
            ]
            result = await asyncio.gather(*tasks)
        return result


    async def request(self, session, url, file, key=None):
        """下载 解密 保存 segment"""

        file = os.path.join(self.store, f"{file}.ts")

        # 本地存在 ts 文件不下载
        if (
            not os.path.isfile(file)
            or not os.path.getsize(file)
        ):
            # 下载 ts 文件
            content = await self.send_request(session=session, url=url)
            # AES 解密
            if key and self.cipher:
                content = self.cipher[key].decrypt(content)
            # 保存 ts 文件
            async with aiofiles.open(file, "wb") as fileobj:
                await fileobj.write(content)
        # 更新进度条
        self.progress.update(self.progress_task, advance=1)
        return file


    @staticmethod
    async def send_request(session, url):
        """请求 segment"""
        while True:
            try:
                async with session.get(url=url) as response:
                    # 异常状态码捕获
                    if 403 <= response.status <= 410:
                        raise aiohttp.InvalidURL(
                            f"invalid url:{str(response.url)!r}, status_code: {response.status!r}"
                        )
                    # 重试部分请求
                    if response.status != 200:
                        await asyncio.sleep(1)
                        continue

                    content = await response.content.read()
                    return content
            # 请求超时 重试
            except asyncio.exceptions.TimeoutError:
                await asyncio.sleep(1)
            # 连接错误 重试
            except (
                aiohttp.client_exceptions.ClientOSError,
                aiohttp.client_exceptions.ClientPayloadError,
                aiohttp.client_exceptions.ClientConnectorError,
            ):
                await asyncio.sleep(2)
            # 服务器拒绝连接
            except aiohttp.client_exceptions.ServerDisconnectedError:
                await asyncio.sleep(2)


    @classmethod
    def start_download_from_segments(
        cls, *, m3u8parse: typing.Union["M3U8Parse"]
    ):
        """类方法 下载segments"""
        self = cls(m3u8parse=m3u8parse)
        try:
            self.start_download()
        finally:
            self.close()
        return self


    def merge_segments(self):
        """合并 segments 文件"""

        # 创建 filelist 文件
        filelist, filesize = os.path.join(self.store, "filelist.txt"), 0
        with utils.open(filelist, "w", encoding="utf-8") as fileobj:
            for file in self.result:
                fileobj.write(f"file '{file}'\n")
                filesize += (os.path.getsize(file) - 16400)
        self.result.append(filelist)

        # ffmpeg 视频合并代码, 监测合并完成状态
        output = self.m3u8parse.m3u.output
        args   = shlex.split(
            f"ffmpeg -loglevel quiet -f concat -safe 0 -i {filelist} -c copy {output} -y"
        )
        merge_completed = False

        # ffmpeg 合并函数
        def merge():
            try:
                subprocess.run(args=args, check=True)
            except FileNotFoundError as error:
                reason = getattr(error, "filename", None)
                if reason != "ffmpeg":
                    raise
                warnings.warn("ffmpeg没在找到, 调用失败", stacklevel=1)
                self.m3u8parse.m3u._reserve = True
            finally:
                nonlocal merge_completed
                merge_completed = True

        # 进度条显
        def show_progress():
            completed_filezise = lambda: os.path.getsize(output) if os.path.isfile(output) else 0
            with progress.Progress(
                *(
                    progress.SpinnerColumn("line"),
                    progress.TextColumn("[progress.description]{task.description}"),
                    progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                    progress.BarColumn(finished_style="green"),
                    # progress.TextColumn("[progress.description]{task.completed}/{task.total}"),
                    progress.DownloadColumn(),
                    # progress.TransferSpeedColumn(),
                    # progress.TimeElapsedColumn()
                ),
                transient=True
            ) as show_progress:
                progress_task = show_progress.add_task(
                    description=os.path.basename(output),
                    total=filesize
                )
                while True:
                    show_progress.update(progress_task, completed=completed_filezise())
                    if merge_completed:
                        show_progress.update(progress_task, completed=filesize)
                        break
                    time.sleep(0.1)

        task1 = threading.Thread(target=merge)
        task2 = threading.Thread(target=show_progress)
        task1.start()
        task2.start()
        task1.join()
        task2.join()



import re, copy
import rich, click
from pyxk import requests

from pyxk.m3u8._m3u8 import load_url, load_content



@click.group(
    invoke_without_command=False,
    chain=True
)
@click.pass_context
@click.option(
    "-o",
    "--output",
    "output",
    type=str,
    default=None,
    help="m3u8 视频保存路径"
)
@click.option(
    "--rerequest",
    is_flag=True,
    default=False,
    help="重新请求 m3u8 网络资源"
)
@click.option(
    "--no-delete",
    "delete",
    is_flag=True,
    default=True,
    help="删除本地 m3u8 文件"
)
@click.option(
    "--reserve",
    "reserve",
    is_flag=True,
    default=False,
    help="保留 segments"
)
@click.option(
    "-h",
    "--headers",
    "headers",
    type=(str, str),
    multiple=True,
    help="Request 请求头"
)
@click.option(
    "--no-verify",
    "verify",
    is_flag=True,
    default=True,
    help="Request SSL验证"
)
@click.option(
    "-l",
    "--limit",
    "limit",
    type=int,
    default=8,
    help="异步下载 Limit"
)
def main(
    ctx, output, rerequest, delete, reserve, headers, verify, limit
):
    """M3U8 下载, 合并管理器(ffmpeg)"""
    ctx.obj = ctx.params


@main.command()
@click.pass_obj
@click.argument(
    "content",
    type=click.Path(exists=True),
    metavar="path <m3u8 文件路径>"
)
@click.option(
    "-u",
    "--url",
    "url",
    type=str,
    default=None,
    help="m3u8 链接地址"
)
@click.option(
    "-o",
    "--output",
    "output",
    type=str,
    default=None,
    help="m3u8 视频保存路径"
)
def file(obj, content, url, output):
    """使用 m3u8文件 下载"""

    kwargs = copy.deepcopy(obj)
    if output is not None:
        kwargs["output"] = output
    load_content(content=content, url=url, **kwargs)


@main.command()
@click.pass_obj
@click.argument(
    "url",
    type=str,
    metavar="url <m3u8 链接地址>"
)
@click.option(
    "-o",
    "--output",
    "output",
    type=str,
    default=None,
    help="m3u8 视频保存路径"
)
def url(obj, url, output):
    """使用 m3u8链接 下载"""

    kwargs = copy.deepcopy(obj)
    if output is not None:
        kwargs["output"] = output
    load_url(url=url, **kwargs)


@main.command()
@click.pass_obj
@click.argument(
    "designation",
    type=str,
    metavar="designation <Jable 番号>"
)
@click.option(
    "-o",
    "--output",
    "output",
    type=str,
    default=None,
    help="m3u8 视频保存路径"
)
def jable(obj, designation, output):
    """Jable.tv 视频下载"""

    kwargs = copy.deepcopy(obj)
    if output is not None:
        kwargs["output"] = output
    jable_video_url = f"https://jable.tv/videos/{designation}/"

    response = requests.get(jable_video_url)
    if response.status_code != 200:
        rich.print(
            f"[bold]Jable[/]: [yellow]{jable_video_url!r}[/] [red]{response.status_code!r}[/]")
        return

    m3u8url = re.search(
        pattern=r"var hlsUrl\s*?=\s*?'(?P<m3u8>.*?)'",
        string=response.text,
        flags=re.M
    )

    if not m3u8url:
        rich.print("[red]没有匹配的M3U8 URL[/]")
        return

    m3u8url = m3u8url.group("m3u8")
    load_url(url=m3u8url, **kwargs)



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyxk",
    version="0.4.4",
    author="xk",
    install_requires=[
        "requests",
        "pycryptodome",
        "rich",
        "m3u8",
        "aiohttp",
        "aiofiles",
        "click",
        "httpx"
    ],
    entry_points={
        'console_scripts': [
            'm3u8 = pyxk.m3u8._entry_point:main',
        ],
    },
    author_email="925330867@qq.com",
    description="pyxk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)


import httpx
from pyxk.lazy_loader import LazyLoader

os = LazyLoader("os", globals())
time = LazyLoader("time", globals())
warnings = LazyLoader("warnings", globals())
console = LazyLoader("console", globals(), "rich.console")
progress = LazyLoader("progress", globals(), "rich.progress")
utils = LazyLoader("utils", globals(), "pyxk.utils")


class Client(httpx.Client):
    """httpx.Client 重构

    pip install httpx
    """

    def __init__(self, headers=None, follow_redirects=True, **kwargs):
        headers = httpx.Headers(headers)
        headers.setdefault("user-agent", utils.UA_ANDROID)
        super().__init__(headers=headers, follow_redirects=follow_redirects, **kwargs)
        self.__console = console.Console()

    def send(self, request, **kwargs):
        exc_count, exc_max = {}, 10
        while True:
            try:
                with self.__console.status(f"[magenta b]Send Request[/]: {str(request.url)}"):
                    response = super().send(request, **kwargs)
                break
            # 请求超时
            except httpx.ConnectTimeout as exc:
                reason  = exc.args[0]
                timeout = request.extensions["timeout"]["connect"]
                warnings.warn(f"timeout: {timeout!r}", stacklevel=1)
                # except 计数
                exc_count.setdefault(reason, 0)
                exc_count[reason] += 1
                if exc_count[reason] >= exc_max:
                    raise
            # 无网络连接
            except httpx.ConnectError as exc:
                reason, reason_re = exc.args[0], ("[Errno 7]", )
                reason_ok = lambda : True in [reason.startswith(item) for item in reason_re]
                if not reason_ok():
                    raise
                warnings.warn("请检查网络连接是否正常...", stacklevel=1)
                time.sleep(1)
        return response

    def head(self, url, **kwargs):
        kwargs.setdefault("follow_redirects", False)
        return super().head(url=url, **kwargs)

    def wget(self, url, method="GET", *, output=None, resume=False, content=None, data=None, files=None, json=None, params=None, headers=None, cookies=None, auth=None, follow_redirects=True, timeout=5, extensions=None):
        """流媒体下载大文件

        :params: output: 下载文件路径
        :params: resume: 断点续传(default: False)
        """
        output = self.__set_wget_output(output, url, params=params, headers=headers, cookies=cookies, auth=auth, timeout=timeout, extensions=extensions)
        file_size, file_mode, chunk_size = 0, "wb", 1024
        # 开启文件续传
        if resume:
            if not output:
                warnings.warn("文件续传缺少 output")
            elif os.path.isfile(output):
                file_size, file_mode = os.path.getsize(output), "ab"
                headers = httpx.Headers(headers)
                headers["range"] = f"bytes={file_size}-"
        # 开启流式响应
        request  = self.build_request(method=method, url=url, content=content, data=data, files=files, json=json, params=params, headers=headers, cookies=cookies, timeout=timeout, extensions=extensions)
        response = self.send(request=request, auth=auth, follow_redirects=follow_redirects, stream=True)
        if not output:
            return response
        content_length = response.headers.get("Content-Length")
        if content_length is not None:
            content_length = int(content_length) + file_size
        with progress.Progress(
            *(
                # progress.SpinnerColumn("line"),
                progress.TextColumn("[progress.description]{task.description}"),
                progress.TaskProgressColumn("[progress.percentage]{task.percentage:>6.2f}%"),
                progress.BarColumn(finished_style="green"),
                progress.DownloadColumn(),
                # progress.TransferSpeedColumn(),
                progress.TimeElapsedColumn()
            ),
            console=self.__console, transient=True
        ) as download_progress:
            description = f"[bold]{os.path.basename(output)}[/]"
            download_task = download_progress.add_task(description=description, total=content_length)
            download_progress.update(download_task, advance=file_size)

            with utils.open(file=output, mode=file_mode) as file_obj:
                for chunk in response.iter_bytes(chunk_size):
                    file_obj.write(chunk)
                    download_progress.update(download_task, advance=chunk_size)
        return response

    def __set_wget_output(self, output, url, *, params=None, headers=None, cookies=None, auth=None, timeout=5, extensions=None):
        if not isinstance(output, str):
            return None
        output = os.path.normpath(os.path.abspath(output))
        if len(os.path.basename(output).rsplit(".", 1)) >= 2:
            return output
        response = self.head(url, params=params, headers=headers, cookies=cookies, auth=auth, timeout=timeout, extensions=extensions)
        content_type = response.headers.get("content-type")
        if not content_type:
            return None
        content_type = content_type.split(";", 1)[0].strip()
        content_type = content_type.rsplit("/", 1)[-1].strip()
        return output + "." + content_type if content_type else output
